/***************************************************************************
	mmsuper8.cpp - Converts PNM scans of super 8 film to MOV

			-------------------
        begin           : 5/31/2004
        copyright       : (C) 2004 by Michael Minn
        email           : see www.michaelminn.com
                        -------------------

    This program is free software; you can redistribute it and/or modify
    it under the terms of version 2 of the GNU General Public License as 
    published by the Free Software Foundation.

    Super 8 dimensions (from http://member.melbpc.org.au/~cksm/Formats.html)
    with pixel dimensions at 1200 DPI:

	Film width: 0.314 in (377px)
	Frame height: 0.1667in (200px)
	Frame width: 0.231 (277px)
	Picture width: 0.228 in
	Picture height: 0.163in
	Left margin (sprocket area): 0.058in (70px)
	Right margin (sound stripe): 0.025in (30px)
	Sprocket hole height: 0.045in (54px)
	Sprocket hole width: 0.036in (43px)
	Sprocket to edge: 0.020in (24px)

	www.heroinewarrior.com/quicktime.php3
	smbclient //toshiba1905/frank -I 192.168.1.1 -U frank
	an invalid sample number was found in the movie

***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include "mmpnm.h"
#include "mmavi.h"
#include "mmquicktime.h"

#define MMSUPER8_VERSION "MMSUPER8 v2004.06.20"

/************************************************************************
		Command Line Interface
************************************************************************/

int print_syntax()
{
	fprintf(stderr, "SYNTAX: mmsuper8 [options] [pnmfile1 pnmfile2 ...]\n");
	fprintf(stderr, "See http:\\\\www.michaelminn.com/linux/mmsuper8 for instructions\n");

	return 0;
}

int main(int argc, char **argv)
{
	int normalize = 0;
	int color_high[3] = { 255, 255, 255 };
	int color_low[3] = { 0, 0, 0 };
	int frames_per_second = 18;
	int color_mode = MMQUICKTIME_COLOR_MODE_RGB;
	char *output_filename = "";
	mmpnm_framing parameters;

	fprintf(stderr, "%s\n", MMSUPER8_VERSION);

	// First pass through arguments for parameters and one-shot actions

	for (int argument = 1; argument < argc; ++argument)
		if (strstr(argv[argument], "-o") && ((argument + 1) < argc))
		{
			++argument;
			output_filename = argv[argument];
		}

		else if (strncmp(argv[argument], "-n", 2) == 0)
			normalize = 1;

		else if (strstr(argv[argument], "-resolution") && ((argument + 1) < argc))
		{
			++argument;
			parameters.set_resolution(atoi(argv[argument]));
		}

		else if ((strncmp(argv[argument], "-l", 2) == 0) && ((argument + 1) < argc))
		{
			++argument;
			if (strstr(argv[argument], ".mov"))
				mmquicktime_list(argv[argument]);
			else if (strstr(argv[argument], ".avi"))
				mmavi_list(argv[argument]);
			else
				fprintf(stderr, "Unrecognized file type for listing: %s\n", argv[argument]);

			return 0;
		}

		else if (strstr(argv[argument], "-rate") && ((argument + 1) < argc))
		{
			++argument;
			frames_per_second = atoi(argv[argument]);
		}

		else if (strcmp(argv[argument], "-b") == 0)
		{
			color_mode = MMQUICKTIME_COLOR_MODE_GRAY;
			fprintf(stderr, "Output will be black and white (grayscale)\n");
		}

		else if (strstr(argv[argument], "-red") && ((argument + 2) < argc))
		{
			color_low[0] = atoi(argv[argument + 1]);
			color_high[0] = atoi(argv[argument + 2]);
			argument += 2;
			fprintf(stderr, "Expanding red between %d and %d\n", color_low[0], color_high[0]);
		}	

		else if (strstr(argv[argument], "-green") && ((argument + 2) < argc))
		{
			color_low[1] = atoi(argv[argument + 1]);
			color_high[1] = atoi(argv[argument + 2]);
			argument += 2;
			fprintf(stderr, "Expanding green between %d and %d\n", color_low[1], color_high[1]);
		}	

		else if (strstr(argv[argument], "-blue") && ((argument + 2) < argc))
		{
			color_low[2] = atoi(argv[argument + 1]);
			color_high[2] = atoi(argv[argument + 2]);
			argument += 2;
			fprintf(stderr, "Expanding blue between %d and %d\n", color_low[2], color_high[2]);
		}	

		else if (strncmp(argv[argument], "-h", 2) == 0)
			return print_syntax();

		else if (strstr(argv[argument], "-clipleft")
			 || strstr(argv[argument], "-rate")
			 || strstr(argv[argument], "-frameheight")
			 || strstr(argv[argument], "-frameoffset")
			 || strstr(argv[argument], "-framenumber")
			 || strstr(argv[argument], "-clipright")
			 || strstr(argv[argument], "-cliptop")
			 || strstr(argv[argument], "-clipbottom"))
		{
			++argument;
			// Just check these in first pass
			if (argument >= argc)
			{
				fprintf(stderr, "%s argument requires a parameter\n", argv[argument]);
				return print_syntax();
			}
		}

		else if (strstr(argv[argument], ".ppm") 
			 || strstr(argv[argument], ".pgm")
			 || strstr(argv[argument], ".pnm")
			 || strstr(argv[argument], "-framefromright")
			 || strstr(argv[argument], "-framefromleft")
			 || strstr(argv[argument], "-nosprocket")
			 || strstr(argv[argument], "-sprocket"))
			continue; // processed below in 2nd pass

		else
		{
			fprintf(stderr, "Invalid argument: %s\n", argv[argument]);
			return print_syntax();
		}

	// Open output file if one is given
	mmavi_file avi_file;
	mmquicktime_file quicktime_file;
	if (strstr(output_filename, ".mov"))
	{
		if (quicktime_file.open_file(output_filename, color_mode, frames_per_second) < 0)
			return -1;
	}
	else if (strstr(output_filename, ".avi"))
	{
		if (avi_file.open_file(output_filename, color_mode, frames_per_second) < 0)
			return -1;
	}
	else if (strlen(output_filename))
	{
		fprintf(stderr, "Output file of unknown type (must be .mov or .avi): %s\n", output_filename);
		return -1;
	}


	// Second pass through arguments to process input files and context sensitive parameters
	int frame_count = 0;
	for (int argument = 1; argument < argc; ++argument)
		if (strstr(argv[argument], "-clipleft"))
		{
			++argument;
			parameters.clip_left = ((atoi(argv[argument]) + 1) & 0xfffffffe);
		}

		else if (strstr(argv[argument], "-clipright"))
		{
			++argument;
			parameters.clip_right = ((atoi(argv[argument]) + 1) & 0xfffffffe);
		}

		else if (strstr(argv[argument], "-cliptop"))
		{
			++argument;
			parameters.clip_top = atoi(argv[argument]);
		}

		else if (strstr(argv[argument], "-clipbottom"))
		{
			++argument;
			parameters.clip_bottom = atoi(argv[argument]);
		}

		else if (strstr(argv[argument], "-framenumber"))
		{
			++argument;
			frame_count = atoi(argv[argument]);
		}
		
		else if (strstr(argv[argument], "-frameheight"))
		{
			++argument;
			parameters.ignore_sprocket_holes = 1;
			parameters.frame_height = atof(argv[argument]);
			if (parameters.frame_height <= 0)
			{
				fprintf(stderr, "Invalid frame height given %f\n", 
					parameters.frame_height);
				return print_syntax();
			}
		}
		
		else if (strstr(argv[argument], "-frameoffset"))
		{
			++argument;
			parameters.ignore_sprocket_holes = 1;
			parameters.frame_offset = atoi(argv[argument]);
			if (parameters.frame_offset < 0)
			{
				fprintf(stderr, "Invalid frame offset given %d\n", parameters.frame_offset);
				return print_syntax();
			}
		}
		
		else if (strstr(argv[argument], "-framefromright"))
			parameters.frame_from_right = 1;
			
		else if (strstr(argv[argument], "-framefromleft"))
			parameters.frame_from_right = 0;
			
		else if (strstr(argv[argument], "-nosprocket"))
			parameters.ignore_sprocket_holes = 1;

		else if (strstr(argv[argument], "-sprocket"))
			parameters.ignore_sprocket_holes = 0;

		else if (strstr(argv[argument], ".ppm") 
			 || strstr(argv[argument], ".pgm")
			 || strstr(argv[argument], ".pnm"))
		{
			mmpnm_file file;
			if (file.load(argv[argument]) >= 0)
			{
				fprintf(stderr, "%s: %d x %d\n", argv[argument], file.width, file.height);

				mmpnm_file *frames = NULL;
				if ((file.width > parameters.film_width) || (file.width < file.height))
					file.convert_to_frames(frames, parameters);

				else // single frame file
				{
					fprintf(stderr, "  Single frame file\n");
					frames = new mmpnm_file;
					frames->copy(&file);
				}

				if (normalize)
					for (mmpnm_file *scan = frames; scan; scan = scan->next)
						scan->normalize();

				for (int color = 0; color < 3; ++color)
					if ((color_low[color] > 0) || (color_high[color] < 255))
						for (mmpnm_file *scan = frames; scan; scan = scan->next)
							scan->adjust_color(color, 
								color_low[color], color_high[color]);
		
				if (!strlen(output_filename))
					frame_count = frames->save_frames_to_separate_files
						(color_mode == MMQUICKTIME_COLOR_MODE_GRAY, frame_count);

				else for (mmpnm_file *scan = frames; scan; scan = scan->next, ++frame_count)
					if (strstr(output_filename, ".mov"))
						quicktime_file.add_frame(scan);
					else
						avi_file.add_frame(scan);

				frames->delete_list();
			}
		}

	if (strstr(output_filename, ".mov"))
		quicktime_file.close_file();

	else if (strstr(output_filename, ".avi"))
		avi_file.close_file();

	fprintf(stderr, "%d frames saved\n", frame_count);

	return 0;
}
