/***************************************************************************
	mmsuper8stabilize - Stabilizes jitter and luminance

			-------------------
        begin           : 6/10/2004
        copyright       : (C) 2004 by Michael Minn
        email           : see www.michaelminn.com
                        -------------------

    This program is free software; you can redistribute it and/or modify
    it under the terms of version 2 of the GNU General Public License as 
    published by the Free Software Foundation.

***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include "mmpnm.h"

int stabilize(int max_shift, int file_count, char **file_list)
{
	int x_shift[file_count];
	int y_shift[file_count];

	// Find the jitter between frames based on correlation
	for (int file = 0; file < (file_count - 1); ++file)
	{
		mmpnm_file file1;
		mmpnm_file file2;

		if (file1.load(file_list[file]) < 0)
			return -1;

		if (file2.load(file_list[file + 1]) < 0)
			return -1;

		if ((file1.width != file2.width) || (file1.height != file2.height))
			return fprintf(stderr, "File dimensions must all be equal: %dx%d vs %dx%d\n",
				file1.width, file1.height, file2.width, file2.height);

		x_shift[file] = 0;
		y_shift[file] = 0;
		unsigned int max_correlation = 0;

		for (int y_shift_scan = -max_shift, entry = 0; y_shift_scan <= max_shift; ++y_shift_scan)
			for (int x_shift_scan = -max_shift; x_shift_scan <= max_shift; ++x_shift_scan, ++entry)
			{
				unsigned int sum = 0;
				for (int y = max_shift; y < (file1.height - max_shift); ++y)
					for (int x = max_shift; x < (file1.width - max_shift); ++x)
						sum += ((unsigned int) file1.gray_data[(y * file1.width) + x]
							* (unsigned int) file2.gray_data
							   [((y + y_shift_scan) * file2.width) 
							    + x + x_shift_scan]);

				if (sum > max_correlation)
				{
					max_correlation = sum;
					x_shift[file] = x_shift_scan;
					y_shift[file] = y_shift_scan;
				}
			}

		printf("Correlating %s and %s (%d x %d); Shift: %d x %d\n", 
			file_list[file], file_list[file + 1], file1.width, file1.height,
			x_shift[file], y_shift[file]);
	}

	// Shift the frames based on deviation from average
	int x_shift_sum = 0;
	int y_shift_sum = 0;
	for (int file = 1; file < file_count; ++file)
	{
		mmpnm_file file1;
		if (file1.load(file_list[file]) < 0)
			return -1;
	
		x_shift_sum -= x_shift[file - 1];
		if (x_shift_sum > max_shift)
			x_shift_sum = max_shift;

		else if (x_shift_sum < -max_shift)
			x_shift_sum = -max_shift;

		y_shift_sum -= y_shift[file - 1];
		if (y_shift_sum > max_shift)
			y_shift_sum = max_shift;

		else if (y_shift_sum < -max_shift)
			y_shift_sum = -max_shift;

		printf("%s: shifting %d x %d\n", file_list[file], x_shift_sum, y_shift_sum);
	
		file1.shift(x_shift_sum, y_shift_sum);

		if (file1.save(file_list[file]) < 0)
			return -1;
	}

	// Find average luminance
	int average_luminance = 0;
	int luminance[file_count];
	for (int file = 0; file < file_count; ++file)
	{
		mmpnm_file file1;
		if (file1.load(file_list[file]) < 0)
			return -1;

		luminance[file] = 0;		
		for (int y = 0, index = 0; y < file1.height; ++y)
			for (int x = 0; x < file1.width; ++x, ++index)
				luminance[file] += file1.gray_data[index];

		luminance[file] /= (file1.width * file1.height);
		average_luminance += luminance[file];
		printf("%s: Luminance %d\n", file_list[file], luminance[file]);
	}

	average_luminance /= file_count;
	printf("Average luminance: %d\n", average_luminance);

	// Modify luminance based on deviation from average
	for (int file = 0; file < file_count; ++file)
	{
		mmpnm_file file1;
		if (file1.load(file_list[file]) < 0)
			return -1;

		printf("%s: luminance correction %f (%d/%d)\n", file_list[file], 
			average_luminance / (double) luminance[file],
			average_luminance, luminance[file]);

		for (int index = 0; (index < (file1.width * file1.height)) && luminance[file]; ++index)
		{
			file1.gray_data[index] = (unsigned char)((int) file1.gray_data[index] 
					* average_luminance / luminance[file]);
			file1.color_data[index * 3] = (unsigned char)((int) file1.color_data[index * 3] 
					* average_luminance / luminance[file]);
			file1.color_data[(index * 3) + 1] = (unsigned char)((int) file1.color_data[(index * 3) + 1] 
					* average_luminance / luminance[file]);
			file1.color_data[(index * 3) + 2] = (unsigned char)((int) file1.color_data[(index * 3) + 2] 
					* average_luminance / luminance[file]);
		}

		if (file1.save(file_list[file]) < 0)
			return -1;
	}

	return 0;
}

/************************************************************************
		Command Line Interface
************************************************************************/

int print_syntax()
{
	fprintf(stderr, "SYNTAX: mmsuper8stabilize <max shift> <pnmfile1> <pnmfile2> ...\n");
	fprintf(stderr, "See http:\\\\www.michaelminn.com/linux/mmsuper8 for instructions\n");

	return 0;
}

int main(int argc, char **argv)
{
	if (argc < 4)
		return print_syntax();

	int maxshift = atoi(argv[1]);
	if (maxshift < 4)
		return print_syntax();

	stabilize(maxshift, argc - 2, &argv[2]);

	return 0;
}
